"""Tests for `ember_mug.connection`."""
from unittest.mock import AsyncMock, Mock, patch

import pytest
from bleak import BleakError

from ember_mug.connection import INITIAL_ATTRS, UPDATE_ATTRS, EmberMugConnection
from ember_mug.consts import (
    TEMP_CELSIUS,
    TEMP_FAHRENHEIT,
    UUID_BATTERY,
    UUID_CONTROL_REGISTER_DATA,
    UUID_CURRENT_TEMPERATURE,
    UUID_DSK,
    UUID_LED,
    UUID_LIQUID_LEVEL,
    UUID_LIQUID_STATE,
    UUID_MUG_ID,
    UUID_MUG_NAME,
    UUID_OTA,
    UUID_TARGET_TEMPERATURE,
    UUID_TEMPERATURE_UNIT,
    UUID_TIME_DATE_AND_ZONE,
    UUID_UDSK,
)
from ember_mug.data import Colour


@patch('ember_mug.connection.USES_BLUEZ', True)
def test_adapter_with_bluez(ember_mug):
    connection = EmberMugConnection(ember_mug, adapter='hci0')
    assert connection._client_kwargs['adapter'] == 'hci0'


@patch('ember_mug.connection.USES_BLUEZ', False)
def test_adapter_without_bluez(ember_mug):
    with pytest.raises(ValueError):
        EmberMugConnection(ember_mug, adapter='hci0')


@patch('ember_mug.connection.EmberMugConnection.update_initial')
@patch('ember_mug.connection.establish_connection')
@pytest.mark.asyncio
async def test_connect(mug_update_initial, mock_establish_connection, mug_connection):
    # Already connected
    mug_connection._client = AsyncMock()
    mug_connection._client.is_connected = True
    async with mug_connection:
        pass
    mock_establish_connection.assert_not_called()

    # Not connected
    mug_connection._client = None
    mug_connection.disconnect = AsyncMock()
    async with mug_connection:
        pass
    mock_establish_connection.assert_called()
    mug_update_initial.assert_called()
    assert mug_connection._client is not None
    mug_connection.disconnect.assert_called()


@patch('ember_mug.connection.establish_connection')
@pytest.mark.asyncio
async def test_connect_error(mock_establish_connection, mug_connection):
    mug_connection._client = None
    mock_establish_connection.side_effect = BleakError
    with pytest.raises(BleakError):
        await mug_connection.ensure_connection()


@pytest.mark.asyncio
async def test_disconnect(mug_connection):
    mug_connection._client = AsyncMock()

    mug_connection._client.is_connected = False
    await mug_connection.disconnect()
    mug_connection._client.disconnect.assert_not_called()

    mug_connection._client.is_connected = True
    await mug_connection.disconnect()
    mug_connection._client.disconnect.assert_called()


@pytest.mark.asyncio
async def test_get_mug_meta(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'Yw====-ABCDEFGHIJ'
    meta = await mug_connection.get_meta()
    assert meta.mug_id == 'c'
    assert meta.serial_number == 'ABCDEFGHIJ'
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_MUG_ID)


@pytest.mark.asyncio
async def test_get_mug_battery(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'5\x01'
    battery = await mug_connection.get_battery()
    assert battery.percent == 53.00
    assert battery.on_charging_base is True
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_BATTERY)


@pytest.mark.asyncio
async def test_get_mug_led_colour(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\xf4\x00\xa1\xff'
    colour = await mug_connection.get_led_colour()
    assert colour.as_hex() == '#f400a1'
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_LED)


@pytest.mark.asyncio
async def test_set_mug_led_colour(mug_connection):
    mug_connection._client.write_gatt_char.side_effect = BleakError
    mug_connection.ensure_connection = AsyncMock()
    # Sadly this is the expected behavior for now
    with pytest.raises(BleakError):
        await mug_connection.set_led_colour(Colour(244, 0, 161))
    mug_connection.ensure_connection.assert_called_once()
    mug_connection._client.write_gatt_char.assert_called_once_with(UUID_LED, bytearray(b'\xf4\x00\xa1\xff'))


@pytest.mark.asyncio
async def test_get_mug_target_temp(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\xcd\x15'
    assert (await mug_connection.get_target_temp()) == 55.81
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_TARGET_TEMPERATURE)


@pytest.mark.asyncio
async def test_set_mug_target_temp(mug_connection):
    mug_connection._client.write_gatt_char.side_effect = BleakError
    mug_connection.ensure_connection = AsyncMock()
    # Sadly this is the expected behavior for now
    with pytest.raises(BleakError):
        await mug_connection.set_target_temp(55.81)
    mug_connection.ensure_connection.assert_called_once()
    mug_connection._client.write_gatt_char.assert_called_once_with(UUID_TARGET_TEMPERATURE, bytearray(b'\xcd\x15'))


@pytest.mark.asyncio
async def test_get_mug_current_temp(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\xcd\x15'
    assert (await mug_connection.get_current_temp()) == 55.81
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_CURRENT_TEMPERATURE)


@pytest.mark.asyncio
async def test_get_mug_liquid_level(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\n'
    assert (await mug_connection.get_liquid_level()) == 10
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_LIQUID_LEVEL)


@pytest.mark.asyncio
async def test_get_mug_liquid_state(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\x06'
    assert (await mug_connection.get_liquid_state()) == 6
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_LIQUID_STATE)


@pytest.mark.asyncio
async def test_get_mug_name(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'Mug Name'
    assert (await mug_connection.get_name()) == 'Mug Name'
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_MUG_NAME)


@pytest.mark.asyncio
async def test_set_mug_name(mug_connection):
    mug_connection._client.write_gatt_char.side_effect = BleakError
    mug_connection.ensure_connection = AsyncMock()
    # Sadly this is the expected behavior for now
    with pytest.raises(ValueError):
        await mug_connection.set_name('Hé!')

    with pytest.raises(BleakError):
        await mug_connection.set_name('Mug name')
    mug_connection.ensure_connection.assert_called()
    mug_connection._client.write_gatt_char.assert_called_once_with(UUID_MUG_NAME, bytearray(b'Mug name'))


@pytest.mark.asyncio
async def test_get_mug_udsk(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'dGVzdCBzdHJpbmc='
    assert (await mug_connection.get_udsk()) == 'test string'
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_UDSK)


@pytest.mark.asyncio
async def test_set_mug_udsk(mug_connection):
    mug_connection._client.write_gatt_char.side_effect = BleakError
    mug_connection.ensure_connection = AsyncMock()
    # Sadly this is the expected behavior for now
    with pytest.raises(BleakError):
        await mug_connection.set_udsk('test string')
    mug_connection.ensure_connection.assert_called_once()
    mug_connection._client.write_gatt_char.assert_called_once_with(UUID_UDSK, bytearray(b'dGVzdCBzdHJpbmc='))


@pytest.mark.asyncio
async def test_get_mug_dsk(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'dGVzdCBzdHJpbmc='
    assert (await mug_connection.get_dsk()) == 'test string'
    mug_connection._client.read_gatt_char.return_value = b'something else'
    assert (await mug_connection.get_dsk()) == "b'something else'"
    mug_connection._client.read_gatt_char.assert_called_with(UUID_DSK)


@pytest.mark.asyncio
async def test_get_mug_temperature_unit(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\x01'
    assert (await mug_connection.get_temperature_unit()) == TEMP_FAHRENHEIT
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_TEMPERATURE_UNIT)
    mug_connection._client.read_gatt_char.reset_mock()
    mug_connection._client.read_gatt_char.return_value = b'\x00'
    assert (await mug_connection.get_temperature_unit()) == TEMP_CELSIUS
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_TEMPERATURE_UNIT)


@pytest.mark.asyncio
async def test_set_mug_temperature_unit(mug_connection):
    mug_connection._client.write_gatt_char.side_effect = BleakError
    mug_connection.ensure_connection = AsyncMock()
    # Sadly this is the expected behavior for now
    with pytest.raises(BleakError):
        await mug_connection.set_temperature_unit(TEMP_CELSIUS)
    mug_connection.ensure_connection.assert_called_once()
    mug_connection._client.write_gatt_char.assert_called_once_with(UUID_TEMPERATURE_UNIT, bytearray(b'\x00'))


@pytest.mark.asyncio
async def test_mug_ensure_correct_unit(mug_connection):
    mug_connection.mug.temperature_unit = TEMP_CELSIUS
    mug_connection.mug.use_metric = True
    mug_connection.set_temperature_unit = AsyncMock(return_value=None)
    await mug_connection.ensure_correct_unit()
    mug_connection.set_temperature_unit.assert_not_called()
    mug_connection.mug.temperature_unit = TEMP_FAHRENHEIT
    await mug_connection.ensure_correct_unit()
    mug_connection.set_temperature_unit.assert_called_with(TEMP_CELSIUS)


@pytest.mark.asyncio
async def test_get_mug_battery_voltage(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'\x01'
    assert (await mug_connection.get_battery_voltage()) == 1
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_CONTROL_REGISTER_DATA)


@pytest.mark.asyncio
async def test_get_mug_date_time_zone(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'c\x0f\xf6\x00'
    date_time = await mug_connection.get_date_time_zone()
    assert date_time.timestamp() == 1661990400.0
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_TIME_DATE_AND_ZONE)


@pytest.mark.asyncio
async def test_read_firmware(mug_connection):
    mug_connection._client.read_gatt_char.return_value = b'c\x01\x80\x00\x12\x00'
    firmware = await mug_connection.get_firmware()
    assert firmware.version == 355
    assert firmware.hardware == 128
    assert firmware.bootloader == 18
    mug_connection._client.read_gatt_char.assert_called_once_with(UUID_OTA)


@pytest.mark.asyncio
async def test_mug_update_initial(mug_connection):
    mug_connection._update_multiple = AsyncMock(return_value={})
    assert (await mug_connection.update_initial()) == {}
    mug_connection._update_multiple.assert_called_once_with(INITIAL_ATTRS)


@pytest.mark.asyncio
async def test_mug_update_all(mug_connection):
    mug_connection._update_multiple = AsyncMock(return_value={})
    assert (await mug_connection.update_all()) == {}
    mug_connection._update_multiple.assert_called_once_with(UPDATE_ATTRS)


@pytest.mark.asyncio
async def test_mug_update_multiple(mug_connection):
    mug_connection.get_name = AsyncMock(return_value='name')
    mug_connection.mug.update_info = AsyncMock()
    await mug_connection._update_multiple(('name',))
    mug_connection.mug.update_info.assert_called_once_with(name='name')


@pytest.mark.asyncio
async def test_mug_update_queued_attributes(mug_connection):
    mug_connection._queued_updates = set()
    assert (await mug_connection.update_queued_attributes()) == []
    mug_connection.get_name = AsyncMock(return_value='name')
    mug_connection.mug.update_info = AsyncMock()
    mug_connection._queued_updates = {'name'}
    await mug_connection.update_queued_attributes()
    mug_connection.mug.update_info.assert_called_once_with(name='name')


def test_mug_notify_callback(mug_connection):
    mug_connection._notify_callback(1, b'\x02')
    mug_connection._notify_callback(1, b'\x02')
    assert mug_connection._latest_event_id == 2
    mug_connection._notify_callback(1, b'\x04')
    assert mug_connection._latest_event_id == 4
    mug_connection._notify_callback(1, b'\x05')
    assert mug_connection._latest_event_id == 5
    mug_connection._notify_callback(1, b'\x06')
    assert mug_connection._latest_event_id == 6
    mug_connection._notify_callback(1, b'\x07')
    assert mug_connection._latest_event_id == 7
    mug_connection._notify_callback(1, b'\x08')
    assert mug_connection._latest_event_id == 8
    callback = Mock()
    mug_connection.register_callback(callback)
    mug_connection._notify_callback(1, b'\x09')
    assert mug_connection._latest_event_id == 9
    callback.assert_not_called()
    assert mug_connection._queued_updates == {
        "battery",
        "target_temp",
        "current_temp",
        "liquid_level",
        "liquid_state",
        "battery_voltage",
    }
    mug_connection._notify_callback(1, b'\x02')
    callback.assert_called_once()



"""Objects and methods related to connection to the mug."""
from __future__ import annotations

import asyncio
import contextlib
import logging
from asyncio import Lock
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Callable

from bleak import BleakClient, BleakError
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak_retry_connector import establish_connection

from .consts import (
    MUG_NAME_REGEX,
    PUSH_EVENT_BATTERY_IDS,
    PUSH_EVENT_ID_AUTH_INFO_NOT_FOUND,
    PUSH_EVENT_ID_BATTERY_VOLTAGE_STATE_CHANGED,
    PUSH_EVENT_ID_CHARGER_CONNECTED,
    PUSH_EVENT_ID_CHARGER_DISCONNECTED,
    PUSH_EVENT_ID_DRINK_TEMPERATURE_CHANGED,
    PUSH_EVENT_ID_LIQUID_LEVEL_CHANGED,
    PUSH_EVENT_ID_LIQUID_STATE_CHANGED,
    PUSH_EVENT_ID_TARGET_TEMPERATURE_CHANGED,
    TEMP_CELSIUS,
    TEMP_FAHRENHEIT,
    USES_BLUEZ,
    UUID_BATTERY,
    UUID_CONTROL_REGISTER_DATA,
    UUID_CURRENT_TEMPERATURE,
    UUID_DSK,
    UUID_LED,
    UUID_LIQUID_LEVEL,
    UUID_LIQUID_STATE,
    UUID_MUG_ID,
    UUID_MUG_NAME,
    UUID_OTA,
    UUID_PUSH_EVENT,
    UUID_TARGET_TEMPERATURE,
    UUID_TEMPERATURE_UNIT,
    UUID_TIME_DATE_AND_ZONE,
    UUID_UDSK,
)
from .data import BatteryInfo, Colour, MugFirmwareInfo, MugMeta
from .utils import bytes_to_big_int, bytes_to_little_int, decode_byte_string, encode_byte_string, temp_from_bytes

if TYPE_CHECKING:
    from .mug import EmberMug

logger = logging.Logger(__name__)

INITIAL_ATTRS = (
    "meta",
    "name",
    "udsk",
    "dsk",
    "date_time_zone",
    "firmware",
)
UPDATE_ATTRS = (
    "led_colour",
    "current_temp",
    "target_temp",
    "temperature_unit",
    "battery",
    "liquid_level",
    "liquid_state",
    "battery_voltage",
)


class EmberMugConnection:
    """Context manager to handle updating via active connection."""

    def __init__(self, mug: EmberMug, adapter: str = None, **kwargs: Any) -> None:
        """Initialize connection manager."""
        self.mug = mug
        self._device: BLEDevice = mug.device
        self._connect_lock: Lock = Lock()
        self._callbacks: list[Callable[[EmberMug], None]] = []
        self._client: BleakClient = None  # type: ignore

        self._queued_updates: set[str] = set()
        self._latest_event_id: int | None = None

        self._client_kwargs = {**kwargs}
        if adapter:
            if USES_BLUEZ is False:
                raise ValueError('The adapter option is only valid for the Linux BlueZ Backend.')
            self._client_kwargs['adapter'] = adapter

    def set_device(self, ble_device: BLEDevice) -> None:
        """Set the ble device."""
        self._device = ble_device

    async def ensure_connection(self) -> None:
        """Connect to mug."""
        if self._client and self._client.is_connected is True:
            return

        async with self._connect_lock:
            # Also check after lock is acquired
            if self._client and self._client.is_connected is True:
                return
            try:
                self._client = await establish_connection(
                    client_class=BleakClient,
                    device=self.mug.device,
                    name=f'{self.mug.name} ({self.mug.device.address})',
                    disconnected_callback=self._disconnect_callback,  # type: ignore
                    ble_device_callback=lambda: self._device,
                    **self._client_kwargs,
                )
            except (asyncio.TimeoutError, BleakError) as error:
                logger.error(f"{self.mug.device}: Failed to connect to the lock: {error}")
                raise error
            # Attempt to pair for good measure and perform an initial update
            with contextlib.suppress(BleakError, EOFError):
                await self._client.pair()
            await self.update_initial()
            await self.subscribe()

    async def disconnect(self) -> None:
        """Disconnect from mug and stop listening to notifications."""
        if self._client and self._client.is_connected is True:
            async with self._connect_lock:
                with contextlib.suppress(BleakError):
                    await self._client.stop_notify(UUID_PUSH_EVENT)
                await self._client.disconnect()

    def _disconnect_callback(self, client: BleakClient) -> None:
        """Disconnect from device."""
        asyncio.create_task(self.disconnect())

    def _fire_callbacks(self) -> None:
        """Fire the callbacks."""
        for callback in self._callbacks:
            callback(self.mug)

    def register_callback(self, callback: Callable[[EmberMug], None]) -> Callable[[], None]:
        """Register a callback to be called when the state changes."""

        def unregister_callback() -> None:
            self._callbacks.remove(callback)

        self._callbacks.append(callback)
        return unregister_callback

    async def __aenter__(self) -> EmberMugConnection:
        """Connect on enter."""
        await self.ensure_connection()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """Cleanup on exit."""
        await self.disconnect()

    async def get_meta(self) -> MugMeta:
        """Fetch Meta info from the mug (Serial number and ID)."""
        return MugMeta.from_bytes(await self._client.read_gatt_char(UUID_MUG_ID))

    async def get_battery(self) -> BatteryInfo:
        """Get Battery percent from mug gatt."""
        return BatteryInfo.from_bytes(await self._client.read_gatt_char(UUID_BATTERY))

    async def get_led_colour(self) -> Colour:
        """Get RGBA colours from mug gatt."""
        return Colour.from_bytes(await self._client.read_gatt_char(UUID_LED))

    async def set_led_colour(self, colour: Colour) -> None:
        """Set new target temp for mug."""
        await self.ensure_connection()
        colour = Colour(*colour[:3], 255)  # It always expects 255 for alpha
        await self._client.write_gatt_char(UUID_LED, colour.as_bytearray())

    async def get_target_temp(self) -> float:
        """Get target temp form mug gatt."""
        temp_bytes = await self._client.read_gatt_char(UUID_TARGET_TEMPERATURE)
        return temp_from_bytes(temp_bytes, self.mug.use_metric)

    async def set_target_temp(self, target_temp: float) -> None:
        """Set new target temp for mug."""
        await self.ensure_connection()
        target = bytearray(int(target_temp / 0.01).to_bytes(2, "little"))
        await self._client.write_gatt_char(UUID_TARGET_TEMPERATURE, target)

    async def get_current_temp(self) -> float:
        """Get current temp from mug gatt."""
        temp_bytes = await self._client.read_gatt_char(UUID_CURRENT_TEMPERATURE)
        return temp_from_bytes(temp_bytes, self.mug.use_metric)

    async def get_liquid_level(self) -> int:
        """Get liquid level from mug gatt."""
        liquid_level_bytes = await self._client.read_gatt_char(UUID_LIQUID_LEVEL)
        return bytes_to_little_int(liquid_level_bytes)

    async def get_liquid_state(self) -> int:
        """Get liquid state from mug gatt."""
        liquid_state_bytes = await self._client.read_gatt_char(UUID_LIQUID_STATE)
        return bytes_to_little_int(liquid_state_bytes)

    async def get_name(self) -> str:
        """Get mug name from gatt."""
        name_bytes: bytearray = await self._client.read_gatt_char(UUID_MUG_NAME)
        return bytes(name_bytes).decode("utf8")

    async def set_name(self, name: str) -> None:
        """Assign new name to mug."""
        await self.ensure_connection()
        if MUG_NAME_REGEX.match(name) is None:
            raise ValueError('Name cannot contain any special characters')
        await self._client.write_gatt_char(UUID_MUG_NAME, bytearray(name.encode("utf8")))

    async def get_udsk(self) -> str:
        """Get mug udsk from gatt."""
        return decode_byte_string(await self._client.read_gatt_char(UUID_UDSK))

    async def set_udsk(self, udsk: str) -> None:
        """Attempt to write udsk."""
        await self.ensure_connection()
        await self._client.write_gatt_char(UUID_UDSK, bytearray(encode_byte_string(udsk)))

    async def get_dsk(self) -> str:
        """Get mug dsk from gatt."""
        value = await self._client.read_gatt_char(UUID_DSK)
        try:
            # TODO: Perhaps it isn't encoded in base64...
            return decode_byte_string(value)
        except ValueError:
            logger.warning("Unable to decode DSK. Falling back to encoded value.")
            return str(value)

    async def get_temperature_unit(self) -> str:
        """Get mug temp unit."""
        unit_bytes = await self._client.read_gatt_char(UUID_TEMPERATURE_UNIT)
        return TEMP_CELSIUS if bytes_to_little_int(unit_bytes) == 0 else TEMP_FAHRENHEIT

    async def set_temperature_unit(self, unit: str) -> None:
        """Set mug unit."""
        await self.ensure_connection()
        unit_bytes = bytearray([1 if unit == TEMP_FAHRENHEIT else 0])
        await self._client.write_gatt_char(UUID_TEMPERATURE_UNIT, unit_bytes)

    async def ensure_correct_unit(self) -> None:
        """Set mug unit if it's not what we want."""
        desired = TEMP_CELSIUS if self.mug.use_metric else TEMP_FAHRENHEIT
        if self.mug.temperature_unit != desired:
            await self.set_temperature_unit(desired)

    async def get_battery_voltage(self) -> int:
        """Get voltage and charge time."""
        battery_voltage_bytes = await self._client.read_gatt_char(UUID_CONTROL_REGISTER_DATA)
        return bytes_to_big_int(battery_voltage_bytes[:1])

    async def get_date_time_zone(self) -> datetime | None:
        """Get date and time zone."""
        date_time_zone_bytes = await self._client.read_gatt_char(UUID_TIME_DATE_AND_ZONE)
        time = bytes_to_big_int(date_time_zone_bytes[:4])
        # offset = bytes_to_big_int(date_time_zone_bytes[4:])
        return datetime.fromtimestamp(time, timezone.utc) if time > 0 else None

    async def get_firmware(self) -> MugFirmwareInfo:
        """Get firmware info."""
        return MugFirmwareInfo.from_bytes(await self._client.read_gatt_char(UUID_OTA))

    async def update_initial(self) -> list[tuple[str, Any, Any]]:
        """Update attributes that don't normally change and don't need to be regularly updated."""
        return await self._update_multiple(INITIAL_ATTRS)

    async def update_all(self) -> list[tuple[str, Any, Any]]:
        """Update all standard attributes."""
        return await self._update_multiple(UPDATE_ATTRS)

    async def _update_multiple(self, attrs: tuple[str, ...]) -> list[tuple[str, Any, Any]]:
        """Helper to update a list of attributes from the mug."""
        changes = self.mug.update_info(**{attr: await getattr(self, f"get_{attr}")() for attr in attrs})
        if changes:
            self._fire_callbacks()
        return changes

    async def update_queued_attributes(self) -> list[tuple[str, Any, Any]]:
        """Update all attributes in queue."""
        if not self._queued_updates:
            return []
        queued_updates = set(self._queued_updates)
        self._queued_updates.clear()
        changes = self.mug.update_info(**{attr: await getattr(self, f"get_{attr}")() for attr in queued_updates})
        if changes:
            self._fire_callbacks()
        return changes

    def _notify_callback(self, characteristic: BleakGATTCharacteristic, data: bytearray):
        """Push events from the mug to indicate changes."""
        event_id = data[0]
        if self._latest_event_id == event_id:
            return  # Skip to avoid unnecessary calls
        logger.info(f"Push event received from Mug ({event_id})")
        self._latest_event_id = event_id

        # Check known IDs
        if event_id in PUSH_EVENT_BATTERY_IDS:
            # 1, 2 and 3 : Battery Change
            if event_id in [
                PUSH_EVENT_ID_CHARGER_CONNECTED,
                PUSH_EVENT_ID_CHARGER_DISCONNECTED,
            ]:
                # 2 -> Placed on charger, 3 -> Removed from charger
                self.on_charging_base = event_id == PUSH_EVENT_ID_CHARGER_CONNECTED
                self._fire_callbacks()
            # All indicate changes in battery
            self._queued_updates.add("battery")
        elif event_id == PUSH_EVENT_ID_TARGET_TEMPERATURE_CHANGED:
            self._queued_updates.add("target_temp")
        elif event_id == PUSH_EVENT_ID_DRINK_TEMPERATURE_CHANGED:
            self._queued_updates.add("current_temp")
        elif event_id == PUSH_EVENT_ID_AUTH_INFO_NOT_FOUND:
            logger.warning("Auth info missing")
        elif event_id == PUSH_EVENT_ID_LIQUID_LEVEL_CHANGED:
            self._queued_updates.add("liquid_level")
        elif event_id == PUSH_EVENT_ID_LIQUID_STATE_CHANGED:
            self._queued_updates.add("liquid_state")
        elif event_id == PUSH_EVENT_ID_BATTERY_VOLTAGE_STATE_CHANGED:
            self._queued_updates.add("battery_voltage")

    async def subscribe(self) -> None:
        """Subscribe to notifications from the mug."""
        try:
            logger.info("Try to subscribe to Push Events")
            await self._client.start_notify(UUID_PUSH_EVENT, self._notify_callback)
        except Exception as e:
            logger.warning(f"Failed to subscribe to state attr {e}")



from unittest.mock import AsyncMock

import pytest
import pytest_asyncio
from bleak.backends.device import BLEDevice

from ember_mug import EmberMug
from ember_mug.connection import EmberMugConnection


@pytest.fixture
def ember_mug():
    yield EmberMug(
        BLEDevice(
            address='32:36:a5:be:88:cb',
            name='Ember Ceramic Mug',
        )
    )


@pytest_asyncio.fixture
async def mug_connection():
    mug = EmberMug(BLEDevice(address='32:36:a5:be:88:cb', name='Ember Ceramic Mug'))
    connection = EmberMugConnection(mug)
    connection._client = AsyncMock()
    yield connection



"""Helpers for formatting values for display."""
from __future__ import annotations

from .consts import LIQUID_STATE_LABELS
from .data import Colour


def format_temp(temp: float, metric: bool = True) -> str:
    """Format temperature with the correct unit."""
    unit = 'C' if metric else 'F'
    return f'{temp:.2f}°{unit}'


def format_led_colour(led_colour: Colour) -> str:
    """Return colour as hex value."""
    return led_colour.as_hex()


def format_liquid_state(liquid_state: int) -> str:
    """Return human-readable liquid state."""
    return LIQUID_STATE_LABELS[liquid_state]


def format_liquid_level(liquid_level: int) -> str:
    """Human readable liquid level."""
    return f'{(liquid_level / 30 * 100):.2f}%'



"""Reusable class for Ember Mug connection and data."""
from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any

from .connection import EmberMugConnection
from .data import BatteryInfo, Colour, MugFirmwareInfo, MugMeta
from .formatting import format_led_colour, format_liquid_level, format_liquid_state, format_temp

if TYPE_CHECKING:
    from bleak.backends.device import BLEDevice


logger = logging.Logger(__name__)

attr_labels = (
    ('name', 'Mug Name'),
    ('meta', 'Meta'),
    ('battery', 'Battery'),
    ('firmware', 'Firmware'),
    ('led_colour', 'LED Colour'),
    ('liquid_state', 'Liquid State'),
    ('liquid_level', 'Liquid Level'),
    ('current_temp', 'Current Temp'),
    ('target_temp', 'Target Temp'),
    ('metric', 'Metric'),
    ('dsk', 'DSK'),
    ('udsk', 'UDSK'),
    ('date_time_zone', 'Date Time + Time Zone'),
    ('battery_voltage', 'Voltage'),
)
extra_attrs = ('dsk', 'udsk', 'battery_voltage', 'date_time_zone')


class EmberMug:
    """Class to connect and communicate with the mug via Bluetooth."""

    name: str = ""
    meta: MugMeta | None = None
    battery: BatteryInfo | None = None
    firmware: MugFirmwareInfo | None = None
    led_colour: Colour = Colour(255, 255, 255)
    liquid_level: int = 0
    liquid_state: int = 0
    current_temp: float = 0.0
    target_temp: float = 0.0
    metric: bool = True
    temperature_unit: str = ""
    dsk: str = ""
    udsk: str = ""
    date_time_zone: str = ""
    battery_voltage: str = ""

    def __init__(self, ble_device: BLEDevice, use_metric: bool = True, include_extra: bool = False) -> None:
        """Set default values in for mug attributes."""
        self.device = ble_device
        self.use_metric = use_metric
        self.model = ble_device.name
        self.include_extra = include_extra

    @property
    def meta_display(self) -> str:
        """Return Meta infor based on preference."""
        if self.meta and not self.include_extra:
            return f'Serial Number: {self.meta.serial_number}'
        return str(self.meta)

    @property
    def led_colour_display(self) -> str:
        """Return colour as hex value."""
        return format_led_colour(self.led_colour)

    @property
    def liquid_state_display(self) -> str:
        """Human-readable liquid state."""
        return format_liquid_state(self.liquid_state)

    @property
    def liquid_level_display(self) -> str:
        """Human-readable liquid level."""
        return format_liquid_level(self.liquid_level)

    @property
    def current_temp_display(self) -> str:
        """Human-readable current temp with unit."""
        return format_temp(self.current_temp, self.use_metric)

    @property
    def target_temp_display(self) -> str:
        """Human-readable target temp with unit."""
        return format_temp(self.target_temp, self.use_metric)

    def update_info(self, **kwargs: Any) -> list[tuple[str, Any, Any]]:
        """Update attributes of the mug if they haven't changed."""
        changes = []
        for attr, new_value in kwargs.items():
            if (old_value := getattr(self, attr)) != new_value:
                setattr(self, attr, new_value)
                changes.append((attr, old_value, new_value))
        return changes

    @property
    def formatted_data(self) -> dict[str, Any]:
        """Return human-readable names and values for all attributes for display."""
        return {
            label: display_value if (display_value := getattr(self, f'{attr}_display', None)) else getattr(self, attr)
            for attr, label in attr_labels
            if self.include_extra or attr not in extra_attrs
        }

    def connection(self, **kwargs: Any) -> EmberMugConnection:
        """Return a connection to the Mug that's meant to be used as a context manager."""
        return EmberMugConnection(self, **kwargs)



"""Tests for `ember_mug.mug`."""
from bleak.backends.device import BLEDevice

from ember_mug.data import BatteryInfo, Colour, MugFirmwareInfo, MugMeta
from ember_mug.mug import EmberMug, EmberMugConnection


def test_mug_formatting():
    mug = EmberMug(BLEDevice(address='32:36:a5:be:88:cb', name='Ember Ceramic Mug'))
    meta = MugMeta('A', 'ABCDEF')
    battery = BatteryInfo(35.46, False)
    firmware = MugFirmwareInfo(12, 35, 55)
    colour = Colour(244, 0, 161)
    mug.update_info(
        name='Mug Name',
        meta=meta,
        battery=battery,
        firmware=firmware,
        led_colour=colour,
        liquid_state=5,
        liquid_level=5,
        current_temp=25,
        target_temp=55,
        metric=True,
        dsk='dsk',
        udsk='udsk',
        date_time_zone=None,
        battery_voltage=0.1,
    )
    assert mug.meta_display == "Serial Number: ABCDEF"
    mug.include_extra = True
    assert mug.meta_display == "Mug ID: A, Serial Number: ABCDEF"
    assert mug.led_colour_display == "#f400a1"
    assert mug.liquid_state_display == "Heating"
    assert mug.liquid_level_display == "16.67%"
    assert mug.current_temp_display == "25.00°C"
    assert mug.target_temp_display == "55.00°C"
    basic_info = {
        'Mug Name': 'Mug Name',
        'Meta': "Serial Number: ABCDEF",
        'Battery': battery,
        'Firmware': firmware,
        'LED Colour': "#f400a1",
        'Liquid State': "Heating",
        'Liquid Level': "16.67%",
        'Current Temp': "25.00°C",
        'Target Temp': "55.00°C",
        'Metric': True,
    }
    assert mug.formatted_data == {
        **basic_info,
        'Meta': "Mug ID: A, Serial Number: ABCDEF",
        'DSK': "dsk",
        'UDSK': "udsk",
        'Date Time + Time Zone': None,
        'Voltage': 0.1,
    }
    mug.include_extra = False
    assert mug.formatted_data == basic_info
    assert isinstance(mug.connection(), EmberMugConnection)



"""Helpers for CLI Interface."""
from __future__ import annotations

import re
from argparse import ArgumentTypeError
from collections import defaultdict
from functools import partial
from typing import TYPE_CHECKING, Any, Callable

from ..consts import MAC_ADDRESS_REGEX
from ..formatting import format_led_colour, format_liquid_level, format_liquid_state, format_temp

if TYPE_CHECKING:
    from ..mug import EmberMug

base_formatters: dict[str, Callable] = {
    'led_colour': format_led_colour,
    'liquid_level': format_liquid_level,
    'liquid_state': format_liquid_state,
}


def validate_mac(value: str) -> str:
    """Check if specified MAC Address is valid."""
    if not isinstance(value, str) or not re.match(MAC_ADDRESS_REGEX, value):
        raise ArgumentTypeError("Invalid MAC Address")
    return value.lower()


def build_sub_rows(row: tuple[str, ...]) -> dict[int, dict[int, str]]:
    """Build a defaultdict of cells to pad for empty values."""
    sub_rows: dict[int, dict[int, str]] = defaultdict(lambda: defaultdict(lambda: ''))
    for i, col in enumerate(row):
        for j, val in enumerate(str(col).split(', ')):
            sub_rows[j][i] = val
    return sub_rows


def print_table(data: list[tuple[str, ...]]) -> None:
    """Print data in a nice little ASCII table."""
    if not data:
        return
    rows = [build_sub_rows(r) for r in data]
    num_columns = max(len(r) for r in rows)
    column_sizes = [max(len(sr[i]) for r in rows for sr in r.values()) + 2 for i in range(num_columns)]
    vertical = f'+{"+".join("-" * i for i in column_sizes)}+'
    print(vertical)
    for row in rows:
        for sub_row in row.values():
            inner = '|'.join(f' {sub_row[i]:<{width-2}} ' for i, width in enumerate(column_sizes))
            print(f'|{inner}|')
        print(vertical)


def print_info(mug: EmberMug) -> None:
    """Print all mug data."""
    print('Mug Data')
    print_table([(k, v) for (k, v) in mug.formatted_data.items()])


def print_changes(changes: list[tuple[str, Any, Any]], metric: bool = True) -> None:
    """Print changes."""
    formatters: dict[str, Callable] = {
        'current_temp': partial(format_temp, metric=metric),
        'target_temp': partial(format_temp, metric=metric),
        **base_formatters,
    }
    for attr, old_value, new_value in changes:
        if formatter := formatters.get(attr):
            old_value, new_value = formatter(old_value), formatter(new_value)
        print(f'{attr.replace("_", " ").title()} changed from {old_value} to {new_value}')



"""Tests for `ember_mug.utils`."""

from ember_mug.utils import (
    bytes_to_big_int,
    bytes_to_little_int,
    decode_byte_string,
    encode_byte_string,
    temp_from_bytes,
)


def test_bytes_to_little_int():
    assert bytes_to_little_int(b'\x05') == 5


def test_bytes_to_big_int():
    assert bytes_to_big_int(b'\x01\xc2') == 450


def test_temp_from_bytes():
    raw_data = bytearray(b'\xcd\x15')  # int: 5581
    assert temp_from_bytes(raw_data) == 55.81
    assert temp_from_bytes(raw_data, metric=False) == 132.46


def test_decode_byte_string():
    assert decode_byte_string(b'dGVzdCBzdHJpbmc=') == 'test string'


def test_encode_byte_string():
    assert encode_byte_string('test string') == b'dGVzdCBzdHJpbmc='



"""Classes for representing data from the mug."""
from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple

from .utils import bytes_to_little_int, decode_byte_string


class Colour(NamedTuple):
    """Simple helper for colour formatting."""

    red: int
    green: int
    blue: int
    alpha: int | None = None

    def as_hex(self) -> str:
        """Format colour array as hex string."""
        return '#' + ''.join(f'{c:02x}' for c in self if c is not None)

    def as_bytearray(self) -> bytearray:
        """Convert to byte array."""
        return bytearray(c for c in self if c is not None)

    @classmethod
    def from_bytes(cls, data: bytes) -> Colour:
        """Initialize from raw bytes."""
        return cls(*data[:3])


@dataclass
class BatteryInfo:
    """Battery Information."""

    percent: float
    on_charging_base: bool

    @classmethod
    def from_bytes(cls, data: bytes) -> BatteryInfo:
        """Initialize from raw bytes."""
        return cls(
            percent=round(float(data[0]), 2),
            on_charging_base=data[1] == 1,
        )

    def __str__(self) -> str:
        """String representation for printing."""
        return f'{self.percent}%, {"" if self.on_charging_base else "not "}on charging base'


@dataclass
class MugFirmwareInfo:
    """Firmware versions."""

    version: int
    hardware: int
    bootloader: int

    @classmethod
    def from_bytes(cls, data: bytes) -> MugFirmwareInfo:
        """Initialize from raw bytes."""
        return cls(
            version=bytes_to_little_int(data[:2]),
            hardware=bytes_to_little_int(data[2:4]),
            bootloader=bytes_to_little_int(data[4:]),
        )

    def __str__(self) -> str:
        """String representation for printing."""
        return ', '.join(
            (
                f'Version: {self.version}',
                f'Hardware: {self.hardware}',
                f'Bootloader: {self.bootloader}',
            )
        )


@dataclass
class MugMeta:
    """Meta data for mug."""

    mug_id: str
    serial_number: str

    @classmethod
    def from_bytes(cls, data: bytes) -> MugMeta:
        """Initialize from raw bytes."""
        return cls(
            mug_id=decode_byte_string(data[:6]),
            serial_number=data[7:].decode("utf8"),
        )

    def __str__(self) -> str:
        """String representation for printing."""
        return f'Mug ID: {self.mug_id}, Serial Number: {self.serial_number}'



"""Tests for `ember_mug.data`."""

from ember_mug.data import BatteryInfo, Colour, MugFirmwareInfo, MugMeta


def test_battery_info():
    battery_info = BatteryInfo.from_bytes(b'5\x01')
    assert battery_info.percent == 53.00
    assert battery_info.on_charging_base is True
    assert str(battery_info) == '53.0%, on charging base'


def test_colour():
    colour = Colour.from_bytes(b'\xf4\x00\xa1\xff')
    assert colour.as_bytearray() == b'\xf4\x00\xa1'
    assert colour.as_hex() == '#f400a1'


def test_mug_firmware_info():
    firmware = MugFirmwareInfo.from_bytes(b'c\x01\x80\x00\x12\x00')
    assert firmware.version == 355
    assert firmware.hardware == 128
    assert firmware.bootloader == 18
    assert str(firmware) == 'Version: 355, Hardware: 128, Bootloader: 18'


def test_mug_meta():
    meta = MugMeta.from_bytes(b'Yw====-ABCDEFGHIJ')
    assert meta.mug_id == 'c'
    assert meta.serial_number == 'ABCDEFGHIJ'
    assert str(meta) == 'Mug ID: c, Serial Number: ABCDEFGHIJ'



"""CLI Interface."""
from __future__ import annotations

import asyncio
import platform
import sys
from argparse import ArgumentParser, Namespace
from typing import TYPE_CHECKING

from bleak import BleakError

from ..mug import EmberMug
from ..scanner import discover_mugs, find_mug
from .helpers import print_changes, print_info, validate_mac

if TYPE_CHECKING:
    from bleak.backends.device import BLEDevice


async def find_device(args: Namespace) -> BLEDevice:
    """Find a single device that has already been paired."""
    try:
        device = await find_mug(mac=args.mac, adapter=args.adapter)
    except BleakError as e:
        print(f'An error occurred trying to find a mug: {e}')
        sys.exit(1)
    if not device:
        print('No mug was found.')
        sys.exit(1)
    print('Found mug:', device)
    return device


async def discover(args: Namespace) -> list[BLEDevice]:
    """Discover new devices in pairing mode."""
    try:
        mugs = await discover_mugs(mac=args.mac)
    except BleakError as e:
        print(f'An error occurred trying to discover mugs: {e}')
        sys.exit(1)
    if not mugs:
        print('No mugs were found. Be sure it is in pairing mode. Or use "find" if already paired.')
        sys.exit(1)

    for mug in mugs:
        print('Found mug', mug)
    return mugs


async def fetch_info(args: Namespace) -> None:
    """Fetch all information from a mug and end."""
    device = await find_device(args)
    mug = EmberMug(device, use_metric=not args.imperial, include_extra=args.extra)
    print('Connecting...')
    async with mug.connection(adapter=args.adapter) as con:
        print('Connected.\nFetching Info')
        await con.update_all()
    print_info(mug)


async def poll_mug(args: Namespace) -> None:
    """Fetch all information and keep polling for changes."""
    device = await find_device(args)
    mug = EmberMug(device, use_metric=not args.imperial, include_extra=args.extra)
    print('Connecting...')
    async with mug.connection(adapter=args.adapter) as con:
        print('Connected.\nFetching Info')
        await con.update_all()
        print_info(mug)
        await con.subscribe()
        print('\nWatching for changes')
        while True:
            for _ in range(60):
                await asyncio.sleep(1)
                print_changes(await con.update_queued_attributes(), con.mug.use_metric)
            # Every minute do a full update
            print_changes(await con.update_all(), con.mug.use_metric)


class EmberMugCli:
    """Very simple CLI Interface to interact with a mug."""

    _commands = {
        'find': find_device,
        'discover': discover,
        'info': fetch_info,
        'poll': poll_mug,
    }

    def __init__(self) -> None:
        """Create parsers."""
        self.parser = ArgumentParser(prog='ember-mug', description='CLI to interact with an Ember Mug')
        shared_parser = ArgumentParser(add_help=False)
        shared_parser.add_argument(
            '-m',
            '--mac',
            action='store',
            type=validate_mac,
            help='Only look for this specific address',
        )
        if platform.system() == 'Linux':
            # Only works on Linux with BlueZ so don't add for others.
            shared_parser.add_argument(
                '-a',
                '--adapter',
                action='store',
                help='Use this Bluetooth adapter instead of the default one (for Bluez)',
            )
        subparsers = self.parser.add_subparsers(dest='command', required=True)
        subparsers.add_parser('find', description='Find the first paired mug', parents=[shared_parser])
        subparsers.add_parser('discover', description='Discover Mugs in pairing mode', parents=[shared_parser])
        info_parsers = ArgumentParser(add_help=False)
        info_parsers.add_argument('-e', '--extra', help='Show extra info', action='store_true')
        info_parsers.add_argument('--imperial', help='Use Imperial units', action='store_true')
        subparsers.add_parser('info', description='Fetch all info from mug', parents=[shared_parser, info_parsers])
        subparsers.add_parser('poll', description='Poll mug for information', parents=[shared_parser, info_parsers])

    async def run(self) -> None:
        """Run the specified command based on subparser."""
        args = self.parser.parse_args()
        if not args.command:
            print('Please specify a command.\n')
            self.parser.print_help()
            sys.exit(1)
        await self._commands[args.command](args)



"""CLI Interface."""
import asyncio

from .commands import EmberMugCli

__all__ = ('EmberMugCli', 'run_cli')


def run_cli() -> None:
    """Run the command line interface."""
    cli = EmberMugCli()
    try:
        asyncio.run(cli.run())
    except KeyboardInterrupt:
        print('Exiting.')



"""Bluetooth UUIDs and other constants used for communicating with the mug."""
from __future__ import annotations

import platform
import re
from os import environ
from uuid import UUID

TEMP_CELSIUS = "C"
TEMP_FAHRENHEIT = "F"

# Bluetooth names of supported mugs
EMBER_BLUETOOTH_NAMES: tuple[str, ...] = ("Ember Ceramic Mug",)

# Name of mug in byte string (Read/Write)
UUID_MUG_NAME = UUID("fc540001-236c-4c94-8fa9-944a3e5353fa")

# Current Mug Temp (Read)
UUID_CURRENT_TEMPERATURE = UUID("fc540002-236c-4c94-8fa9-944a3e5353fa")

# Target Mug Temp (Read/Write)
UUID_TARGET_TEMPERATURE = UUID("fc540003-236c-4c94-8fa9-944a3e5353fa")

# Unit (0 -> Celsius, 1 -> Fahrenheit) (Read/Write)
UUID_TEMPERATURE_UNIT = UUID("fc540004-236c-4c94-8fa9-944a3e5353fa")

# Level (Between 0 -> 30 ?) 30 100% ?
UUID_LIQUID_LEVEL = UUID("fc540005-236c-4c94-8fa9-944a3e5353fa")

# Battery Info (Read)
UUID_BATTERY = UUID("fc540007-236c-4c94-8fa9-944a3e5353fa")

# Integer representing what it is doing with the liquid (Read)
UUID_LIQUID_STATE = UUID("fc540008-236c-4c94-8fa9-944a3e5353fa")

# Constants for liquid state codes
LIQUID_STATE_UNKNOWN = 0
LIQUID_STATE_EMPTY = 1
LIQUID_STATE_FILLING = 2
LIQUID_STATE_COLD_NO_TEMP_CONTROL = 3
LIQUID_STATE_COOLING = 4
LIQUID_STATE_HEATING = 5
LIQUID_STATE_TARGET_TEMPERATURE = 6
LIQUID_STATE_WARM_NO_TEMP_CONTROL = 7

# Labels so liquid states
LIQUID_STATE_LABELS: dict[int, str] = {
    LIQUID_STATE_UNKNOWN: "Unknown",
    LIQUID_STATE_EMPTY: "Empty",
    LIQUID_STATE_FILLING: "Filling",
    LIQUID_STATE_COLD_NO_TEMP_CONTROL: "Cold (No control)",
    LIQUID_STATE_COOLING: "Cooling",
    LIQUID_STATE_HEATING: "Heating",
    LIQUID_STATE_TARGET_TEMPERATURE: "Perfect",
    LIQUID_STATE_WARM_NO_TEMP_CONTROL: "Warm (No control)",
}

EMBER_SERVICE_UUID = UUID("fc543622-236c-4c94-8fa9-944a3e5353fa")

# [Unique ID]-[serial number] (Read)
UUID_MUG_ID = UUID("fc54000d-236c-4c94-8fa9-944a3e5353fa")

# DSK - Unique ID used for auth in app (Read)
UUID_DSK = UUID("fc54000e-236c-4c94-8fa9-944a3e5353fa")

# UDSK - Used for auth in app (Read/Write)
UUID_UDSK = UUID("fc54000f-236c-4c94-8fa9-944a3e5353fa")

# TO watch for changes from mug (Notify/Read)
UUID_PUSH_EVENT = UUID("fc540012-236c-4c94-8fa9-944a3e5353fa")

# Push event codes
PUSH_EVENT_ID_BATTERY_CHANGED = 1
PUSH_EVENT_ID_CHARGER_CONNECTED = 2
PUSH_EVENT_ID_CHARGER_DISCONNECTED = 3
PUSH_EVENT_ID_TARGET_TEMPERATURE_CHANGED = 4
PUSH_EVENT_ID_DRINK_TEMPERATURE_CHANGED = 5
PUSH_EVENT_ID_AUTH_INFO_NOT_FOUND = 6
PUSH_EVENT_ID_LIQUID_LEVEL_CHANGED = 7
PUSH_EVENT_ID_LIQUID_STATE_CHANGED = 8
PUSH_EVENT_ID_BATTERY_VOLTAGE_STATE_CHANGED = 9

PUSH_EVENT_BATTERY_IDS = [
    PUSH_EVENT_ID_BATTERY_CHANGED,
    PUSH_EVENT_ID_CHARGER_CONNECTED,
    PUSH_EVENT_ID_CHARGER_DISCONNECTED,
]

# To gather bytes from mug for stats (Notify)
UUID_STATISTICS = UUID("fc540013-236c-4c94-8fa9-944a3e5353fa")

# RGBA Colour of LED (Read/Write)
UUID_LED = UUID("fc540014-236c-4c94-8fa9-944a3e5353fa")

# Date/Time (Read/Write)
UUID_TIME_DATE_AND_ZONE = UUID("fc540006-236c-4c94-8fa9-944a3e5353fa")

# Last location - (Write)
UUID_LAST_LOCATION = UUID("fc54000a-236c-4c94-8fa9-944a3e5353fa")

# Firmware info (Read)
UUID_OTA = UUID("fc54000c-236c-4c94-8fa9-944a3e5353fa")

# int/temp lock - Address (Read/Write)
UUID_CONTROL_REGISTER_ADDRESS = UUID("fc540010-236c-4c94-8fa9-944a3e5353fa")

# Battery charge info (Read/Write)
UUID_CONTROL_REGISTER_DATA = UUID("fc540011-236c-4c94-8fa9-944a3e5353fa")

# These UUIDs are currently unused. Not for this mug?
UUID_VOLUME = UUID("fc540009-236c-4c94-8fa9-944a3e5353fa")
UUID_ACCELERATION = UUID("fc54000b-236c-4c94-8fa9-944a3e5353fa")

# Validation
MUG_NAME_REGEX = re.compile(r"^[A-Za-z0-9,.\[\]#()!\"\';:|\-_+<>%= ]{1,16}$")
MAC_ADDRESS_REGEX = re.compile(r"^([0-9A-Fa-f]{2}:){5}([0-9A-Fa-f]{2})$")

# Modes
USES_BLUEZ = not environ.get("P4A_BOOTSTRAP") and platform.system() == "Linux"



"""Scanning tools for finding mugs."""
from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Callable

from bleak import BleakScanner

from .consts import EMBER_BLUETOOTH_NAMES, EMBER_SERVICE_UUID, USES_BLUEZ

if TYPE_CHECKING:
    from bleak.backends.device import BLEDevice
    from bleak.backends.scanner import AdvertisementData


logger = logging.Logger(__name__)


def build_scanner_kwargs(adapter: str = None) -> dict[str, Any]:
    """Add Adapter to kwargs for scanner if specified and using BlueZ."""
    if adapter and not USES_BLUEZ:
        raise ValueError('The adapter option is only valid for the Linux BlueZ Backend.')
    return {'adapter': adapter} if adapter else {}


async def discover_mugs(mac: str = None, adapter: str = None, wait: int = 5) -> list[BLEDevice]:
    """Discover new mugs in pairing mode."""
    scanner_kwargs = build_scanner_kwargs(adapter)
    async with BleakScanner(service_uuids=[str(EMBER_SERVICE_UUID)], **scanner_kwargs) as scanner:
        await asyncio.sleep(wait)
        if mac:
            mac = mac.lower()
            return [d for d in scanner.discovered_devices if d.address.lower() == mac]
        return scanner.discovered_devices


def build_find_filter(mac: str = None) -> Callable:
    """Create a filter for finding the mug by name and or mac address."""
    known_names = [n.lower() for n in EMBER_BLUETOOTH_NAMES]

    def mug_filter(device: BLEDevice, advertisement: AdvertisementData) -> bool:
        """Filter by mac if specified else just check the name."""
        if mac is not None and device.address.lower() != mac:
            return False
        return device.name.lower() in known_names

    return mug_filter


async def find_mug(mac: str = None, adapter: str = None) -> BLEDevice | None:
    """Find a mug."""
    if mac is not None:
        mac = mac.lower()
    scanner_kwargs = build_scanner_kwargs(adapter)
    return await BleakScanner.find_device_by_filter(build_find_filter(mac), **scanner_kwargs)



from unittest.mock import patch

import pytest
from bleak.backends.device import BLEDevice

from ember_mug.scanner import build_find_filter, build_scanner_kwargs, discover_mugs, find_mug

MUG_1 = BLEDevice(address='32:36:a5:be:88:cb', name='Ember Ceramic Mug')
MUG_2 = BLEDevice(address='9c:da:8c:19:27:da', name='Ember Ceramic Mug')
EXAMPLE_MUGS = [MUG_1, MUG_2]


@patch('ember_mug.scanner.USES_BLUEZ', True)
def test_build_scanner_kwargs_linux():
    assert build_scanner_kwargs() == {}
    assert build_scanner_kwargs(adapter='hci0') == {'adapter': 'hci0'}


@patch('ember_mug.scanner.USES_BLUEZ', False)
def test_build_scanner_kwargs_other():
    with pytest.raises(ValueError):
        assert build_scanner_kwargs(adapter='hci0')


@patch('asyncio.sleep')
@patch("ember_mug.scanner.BleakScanner")
@pytest.mark.asyncio
async def test_discover_mugs(mock_scanner, mock_sleep):
    mock_scanner.return_value.__aenter__.return_value.discovered_devices = EXAMPLE_MUGS
    mugs = await discover_mugs()
    assert len(mugs) == 2
    mugs = await discover_mugs(mac='32:36:a5:be:88:cb')
    assert len(mugs) == 1
    assert mugs[0].address == '32:36:a5:be:88:cb'
    assert mock_sleep.called_with(delay=5)


@pytest.mark.asyncio
@patch(
    "bleak.BleakScanner.find_device_by_filter",
    return_value=MUG_1,
)
async def test_find_mug(mock_find_device_by_filter):
    # Without filter
    mug = await find_mug()
    assert mug.name == "Ember Ceramic Mug"
    assert mug.address == '32:36:a5:be:88:cb'
    # With Filter
    mug = await find_mug(mac='32:36:a5:be:88:cb')
    assert mug.name == "Ember Ceramic Mug"
    assert mug.address == '32:36:a5:be:88:cb'
    mock_find_device_by_filter.assert_called()


def test_build_find_filter():
    mac_filter = build_find_filter(mac='32:36:a5:be:88:cb')
    assert mac_filter(MUG_1, None) is True
    assert mac_filter(MUG_2, None) is False



from argparse import ArgumentTypeError
from textwrap import dedent

import pytest

from ember_mug.cli.helpers import build_sub_rows, print_changes, validate_mac


def test_validate_mac():
    with pytest.raises(ArgumentTypeError):
        validate_mac('potato')
    assert validate_mac('9C:DA:8C:19:27:DA') == '9c:da:8c:19:27:da'


def test_build_sub_rows():
    sub_rows = build_sub_rows(('Test', 'test1, test2, test3', 'test4'))
    assert sub_rows[0][0] == 'Test'
    assert sub_rows[0][1] == 'test1'
    assert sub_rows[0][2] == 'test4'
    assert sub_rows[1][0] == ''
    assert sub_rows[1][1] == 'test2'
    assert sub_rows[1][2] == ''
    assert sub_rows[2][1] == 'test3'


def test_print_changes(capsys):
    print_changes(
        [
            ('current_temp', 55.1, 25),
            ('target_temp', 50, 25),
        ],
        True,
    )
    captured = capsys.readouterr()
    assert captured.out == dedent(
        """\
        Current Temp changed from 55.10°C to 25.00°C
        Target Temp changed from 50.00°C to 25.00°C
        """
    )



"""Helpful utils for processing mug data."""
from __future__ import annotations

import base64
import re


def decode_byte_string(data: bytes | bytearray) -> str:
    """Convert bytes to text as Ember expects."""
    return base64.decodebytes(data + b"===").decode("utf-8")


def encode_byte_string(data: str) -> bytes:
    """Encode string from Ember Mug."""
    return re.sub(b"[\r\n]", b"", base64.encodebytes(data.encode("utf8")))


def bytes_to_little_int(data: bytearray | bytes) -> int:
    """Convert bytes to little int."""
    return int.from_bytes(data, byteorder="little", signed=False)


def bytes_to_big_int(data: bytearray | bytes) -> int:
    """Convert bytes to big int."""
    return int.from_bytes(data, "big")


def temp_from_bytes(temp_bytes: bytearray, metric: bool = True) -> float:
    """Get temperature from bytearray and convert to fahrenheit if needed."""
    temp = float(bytes_to_little_int(temp_bytes)) * 0.01
    if metric is False:
        # Convert to fahrenheit
        temp = (temp * 9 / 5) + 32
    return round(temp, 2)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ember_mug', 'ember_mug.cli', 'tests', 'tests.cli']

package_data = \
{'': ['*']}

install_requires = \
['bleak-retry-connector>=1.17.1', 'bleak>=0.17.0']

entry_points = \
{'console_scripts': ['ember-mug = ember_mug.cli:run_cli']}

setup_kwargs = {
    'name': 'python-ember-mug',
    'version': '0.3.0',
    'description': 'Python Library for Ember Mugs.',
    'long_description': "# Python Ember Mug\n\n[![pypi](https://img.shields.io/pypi/v/python-ember-mug.svg)](https://pypi.org/project/python-ember-mug/)\n[![python](https://img.shields.io/pypi/pyversions/python-ember-mug.svg)](https://pypi.org/project/python-ember-mug/)\n[![Build Status](https://github.com/sopelj/python-ember-mug/actions/workflows/dev.yml/badge.svg)](https://github.com/sopelj/python-ember-mug/actions/workflows/dev.yml)\n[![codecov](https://codecov.io/gh/sopelj/python-ember-mug/branch/main/graphs/badge.svg)](https://codecov.io/github/sopelj/python-ember-mug)\n\nPython Library for Ember Mugs\n\n* Documentation: <https://sopelj.github.io/python-ember-mug>\n* GitHub: <https://github.com/sopelj/python-ember-mug>\n* PyPI: <https://pypi.org/project/python-ember-mug/>\n* Free software: MIT\n\n## Summary\n\nLibrary to attempt to interact with Ember Mugs via Bluetooth using the bleak library.\nThis was created for use with my [Home Assistant integration](https://github.com/sopelj/hass-ember-mug-component),\nbut could be useful separately and has a simple CLI interface too.\n\n**Note**: I have only tested with my Ember Mug 2, but others should work. (Please let me know)\n\n## Features\n\n* Finding mugs\n* Connecting to Mugs\n* Reading Information (Colour, temp, liquid level, etc.)\n* Polling for changes\n\n## Usage\n\n### Python\n\n```python\nfrom ember_mug.scanner import find_mug, discover_mugs\nfrom ember_mug.mug import EmberMug\n\n# if first time with mug in pairing\nmugs = await discover_mugs()\ndevice = mugs[0]\n# after paired you can simply use\ndevice = await find_mug()\nmug = EmberMug(device)\nasync with mug.connection() as con:\n    print('Connected.\\nFetching Info')\n    await con.update_all()\n    print(mug.formatted_data)\n```\n\n### CLI\n\nIt can also be run via command line either directly with `ember-mug --help` or as a module with `python -m ember_mug --help`\nThere are four options with different subsections. You can see them by specifying them before help. eg `ember-mug poll --help`\n\n```bash\nember-mug discover  # Finds the mug in pairing mode for the first time\nember-mug poll  # fetches info and keeps listening for notifications\n```\n\nBasic options:\n\n| Command    | Use                                                                            |\n|------------|--------------------------------------------------------------------------------|\n| `discover` | Find/List all detected unpaired mugs in pairing mode                           |\n| `find`     | Find *one* already paired mugs                                                 |\n| `info`     | Connect to *one* mug and print its current state                               |\n| `poll`     | Connect to *one* mug and print its current state and keep watching for changes |\n\n![CLI Example](./docs/images/cli-example.png)\n\n## Caveats\n\n- Since this api is not public, a lot of guesswork and reverse engineering is involved, so it's not perfect.\n- Only works with one mug at a time\n- These mugs do not broadcast data unless paired. So you can only have one device connected to it. You need to reset them to change to another device and make sure the previous device doesn't try to reconnect.\n- Reading data from the mug seems to work pretty well, but I have been unable to write to it so far... I always get NotPermitted errors.\n- I haven't figured out some attributes like udsk, dsk, location,\n\n## Credits\n\nThis package was created with [Cookiecutter](https://github.com/audreyr/cookiecutter) and the [waynerv/cookiecutter-pypackage](https://github.com/waynerv/cookiecutter-pypackage) project template.\n",
    'author': 'Jesse Sopel',
    'author_email': 'jesse.sopel@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sopelj/python-ember-mug',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Tests for `ember_mug.formatting`."""

from ember_mug.data import Colour
from ember_mug.formatting import format_led_colour, format_liquid_level, format_liquid_state, format_temp


def test_format_led_colour():
    assert format_led_colour(Colour(244, 0, 161)) == "#f400a1"


def test_format_liquid_level():
    assert format_liquid_level(5) == "16.67%"
    assert format_liquid_level(6) == "20.00%"


def test_format_liquid_state():
    assert format_liquid_state(0) == "Unknown"
    assert format_liquid_state(1) == "Empty"
    assert format_liquid_state(2) == "Filling"
    assert format_liquid_state(3) == "Cold (No control)"
    assert format_liquid_state(4) == "Cooling"
    assert format_liquid_state(5) == "Heating"
    assert format_liquid_state(6) == "Perfect"
    assert format_liquid_state(7) == "Warm (No control)"


def test_format_temp():
    assert format_temp(25.445) == '25.45°C'
    assert format_temp(36.443, metric=False) == '36.44°F'
